"""
This module contains the GUI components for the AxoDen application.

Start the tkinter GUI by running the following command:

.. code-block:: bash

    python -m axoden.gui.gui_projections_quantification

For the streamlit app, run the following command:

.. code-block:: bash

    python -m axoden.gui.streamlit_app
"""



# Import python libraries
import os
import sys
import tkinter as tk
from tkinter import filedialog

if os.getcwd() not in sys.path:
    sys.path.append(os.getcwd())

from axoden.volume_projections import (
    load_table,
    process_folder,
    write_signal_intensity_along_axis_plot,
    write_summary_data_plot,
)


# Function to get the pixel size entered by the user
def get_pixel_size():
    pixel_size = float(pixel_size_entry.get())
    return pixel_size


def update_status(hndl, txt):
    hndl.config(text=txt)
    hndl.update()


def ask_folder():
    folder_path = filedialog.askdirectory()
    folder_path_entry.delete(0, tk.END)
    folder_path_entry.insert(tk.END, folder_path)

    if output_folder_path_entry.get() == "":
        output_folder_path_entry.delete(0, tk.END)
        output_folder_path_entry.insert(
            tk.END, os.path.join(folder_path, "AxoDen_output")
        )


def open_summary_csv_files(folderpath, statuslabel, endswith):
    # Check if there is a csv file in the folder path
    csv_files = [file for file in os.listdir(folderpath) if file.endswith(endswith)]
    if len(csv_files) == 0:
        update_status(
            statuslabel, "\n\nWarning:\nNo csv file found in the folder path."
        )
        return
    elif len(csv_files) > 1:
        update_status(
            statuslabel, "\n\nWarning:\nMultiple csv files found in the folder path."
        )
        return
    else:
        # Read the csv file and plot the data
        csv_file_path = os.path.join(folderpath, csv_files[0])
        table_data = load_table(csv_file_path)

    return table_data


def run_volume_projections():
    """Run the volume projections with the provided folder path and options."""

    # Get the folder path and options
    input_folder_path = folder_path_entry.get()
    output_folder_path = output_folder_path_entry.get()
    data_collection = data_collection_var.get()
    summary_data = summary_data_var.get()
    summary_axes = summary_axes_var.get()
    is_masked = is_masked_var.get()
    pixel_size = get_pixel_size()

    # Check if folder_path is empty
    if not input_folder_path:
        update_status(status_label, "\n\nWarning:\nFolder path is empty or wrong.")
        return

    if output_folder_path == "":
        output_folder_path = input_folder_path

    # Disable the run button
    run_button.config(state=tk.DISABLED)

    # Check if ONLY data collection checkbox is toggled
    if data_collection and not summary_data and not summary_axes:
        update_status(status_label, "\n\nCollecting data...")
        _, _ = process_folder(
            input_folder_path, pixel_size, is_masked, output_folder=output_folder_path
        )
        update_status(
            status_label,
            "\n\nFinished collecting and saving data.\nCheck the control plots!",
        )

    # Check if ONLY summary data checkbox is toggled
    elif summary_data and not data_collection and not summary_axes:
        # Collect and read summary csv file
        table_data = open_summary_csv_files(
            output_folder_path, status_label, "quantification.csv"
        )
        # Plot the data using the table
        update_status(status_label, "\n\nPlotting Summary Data...")
        write_summary_data_plot(output_folder_path, table_data)
        update_status(status_label, "\n\nDone")

    # Check if ONLY summary axis checkbox is toggled
    elif summary_axes and not data_collection and not summary_data:
        # Collect and read summary csv file
        table_data = open_summary_csv_files(
            output_folder_path, status_label, "axis.csv"
        )
        # Plot the data using the table
        update_status(status_label, "\n\nPlotting Summary Axis Data...")
        write_signal_intensity_along_axis_plot(
            output_folder_path, table_data, pixel_size
        )
        update_status(status_label, "\n\nDone")

    # Check if summary data and summary axes checkboxes are toggled
    elif summary_data and summary_axes and not data_collection:
        # Collect and read summary csv files
        table_data = open_summary_csv_files(
            output_folder_path, status_label, "quantification.csv"
        )
        table_data_axes = open_summary_csv_files(
            output_folder_path, status_label, "axis.csv"
        )
        # Plot the data using the tables
        update_status(status_label, "\n\nPlotting Summary Data...")
        write_summary_data_plot(output_folder_path, table_data)
        update_status(
            status_label, "\n\nPlotted Summary Data\nPlotting Summary Axis Data..."
        )
        write_signal_intensity_along_axis_plot(
            output_folder_path, table_data_axes, pixel_size
        )
        update_status(status_label, "\n\nDone")

    # Check if both data collection and summary data checkboxes are toggled
    elif summary_data and data_collection and not summary_axes:
        update_status(status_label, "\n\nCollecting data...")
        table_data, _ = process_folder(
            input_folder_path, pixel_size, is_masked, output_folder=output_folder_path
        )
        update_status(status_label, "\n\nPlotting Summary Data...")
        write_summary_data_plot(output_folder_path, table_data)
        update_status(status_label, "\n\nDone")

    # Check if summary axes checkbox and summary data checkboxes are toggled
    elif summary_axes and data_collection and not summary_data:
        update_status(status_label, "\n\nCollecting data...")
        _, table_data = process_folder(
            input_folder_path, pixel_size, is_masked, output_folder=output_folder_path
        )
        update_status(status_label, "\n\nPlotting Summary Axis Data...")
        write_signal_intensity_along_axis_plot(
            output_folder_path, table_data, pixel_size
        )
        update_status(status_label, "\n\nDone")

    # Check if ALL checkboxes are toggled
    elif data_collection and summary_axes and summary_data:
        update_status(status_label, "\n\nCollecting data...")
        table_data, table_data_axes = process_folder(
            input_folder_path, pixel_size, is_masked, output_folder=output_folder_path
        )
        update_status(status_label, "\n\nPlotting Summary Data...")
        write_summary_data_plot(output_folder_path, table_data)
        update_status(
            status_label, "\n\nPlotted Summary Data\nPlotting Summary Axis Data..."
        )
        write_signal_intensity_along_axis_plot(
            output_folder_path, table_data_axes, pixel_size
        )
        update_status(status_label, "\n\nDone")

    # Enable the run button
    run_button.config(state=tk.NORMAL)


# Function to close the window
def close_window():
    window.destroy()


# Create the main window
window = tk.Tk()
window.title("AxoDen")

# Configure the window size
# window.geometry("300x700")  # Set the width and height of the window
# window.minsize(900, 1600)  # Set the width and height of the window

# Create the pixel size input
pixel_size_label = tk.Label(window, text="\n\nPixel Size (um):")
pixel_size_label.pack()
pixel_size_entry = tk.Entry(window)
pixel_size_entry.insert(tk.END, "0.75521")  # Set the default pixel size
pixel_size_entry.pack()

# Create note for user
note_label = tk.Label(
    window,
    text=(
        "\nNote:\nThe default pixel size is for the 20x Objective.\n"
        "Please change it according to the objective used.\n\nx4 Objective: 3.77396\n"
    ),
)
note_label.pack()

# Call the get_pixel_size function to retrieve the pixel size entered by the user
pixel_size = get_pixel_size()

# Create the input folder path input
folder_path_label = tk.Label(window, text="Input Folder:")
folder_path_label.pack()
folder_path_entry = tk.Entry(window, width=40)
folder_path_entry.pack(padx=20)

# Create the folder path input
ask_folder_button = tk.Button(window, text="Select Folder", command=ask_folder)
ask_folder_button.pack(pady=20)

# Create the output folder path input
output_folder_path_label = tk.Label(window, text="Output Folder:")
output_folder_path_label.pack()
output_folder_path_entry = tk.Entry(window, width=40)
output_folder_path_entry.pack()

# Create the masked option
is_masked_label = tk.Label(
    window,
    text=(
        "\nCheck the box\nif images are masked and have\n"
        "values of 0 where there is no signal:"
    ),
)
is_masked_label.pack()
is_masked_var = tk.BooleanVar(value=True)
is_masked_checkbox = tk.Checkbutton(window, text="", variable=is_masked_var)
is_masked_checkbox.pack()

# Guide the user to select the options
instructions_label = tk.Label(
    window, text="\nCheck the following boxes\naccording to what you want to do:"
)
instructions_label.pack()

# Create the data collection option
data_collection_var = tk.BooleanVar(value=True)
data_collection_checkbox = tk.Checkbutton(
    window, text="Data collection", variable=data_collection_var
)
data_collection_checkbox.pack()

# Create the summary data option
summary_data_var = tk.BooleanVar(value=True)
summary_data_var_checkbox = tk.Checkbutton(
    window, text=" Plot Summary Data", variable=summary_data_var
)
summary_data_var_checkbox.pack()

# Create the summary axes option
summary_axes_var = tk.BooleanVar(value=True)
summary_axes_var_checkbox = tk.Checkbutton(
    window, text="Plot Summary Axis Data", variable=summary_axes_var
)
summary_axes_var_checkbox.pack()

# Create some space between the checkbox and the button
space_label = tk.Label(window, text="\n")
space_label.pack()

# Create the run button
run_button = tk.Button(
    window, text="Run Volume Projections", command=run_volume_projections
)
run_button.pack()

# Create label to update the status of the program
status_label = tk.Label(window, text="\n")
status_label.pack()

# Create the close button
close_button = tk.Button(window, text="Close", command=close_window)
# close_button.place(relx=0.5, rely=1.0, anchor=tk.S, y=-20)
close_button.pack(pady=20)

# Start the GUI event loop
window.mainloop()



"""
axoden simplifies the quantification of axonal projections in neuroscience.
"""

__version__ = "0.3.0"

from .volume_projections import (  # noqa: F401
    load_table,
    plot_signal_intensity_along_axis,
    plot_summary_data,
    process_folder,
    process_image,
    save_table,
    write_signal_intensity_along_axis_plot,
    write_summary_data_plot,
)

__all__ = [
    "load_table",
    "plot_signal_intensity_along_axis",
    "plot_summary_data",
    "process_folder",
    "process_image",
    "save_table",
    "write_signal_intensity_along_axis_plot",
    "write_summary_data_plot",
]



import logging
from io import BytesIO
from typing import Dict, Iterable, List, Tuple

import matplotlib
import matplotlib.pyplot as plt
import numpy as np
import pandas as pd
import pypdf
import streamlit as st
from matplotlib.figure import Figure
from PIL import UnidentifiedImageError
from streamlit.runtime.uploaded_file_manager import UploadedFile

from axoden.gui.streamlit_app.pdf_utils import fig2pdfpage, fig2stream, pages2pdf
from axoden.volume_projections import (
    collect_info_from_filename,
    plot_signal_intensity_along_axis,
    plot_summary_data,
    process_image,
)

matplotlib.use("Agg")  # fixes hanging tests due to use of tk backend

logger = logging.getLogger(__name__)
logging.basicConfig(level=logging.WARNING)


def init_session_state():
    """Initializes the session state variables."""
    if "figures" not in st.session_state:
        st.session_state.figures = []

    if "ctrl_plots_pdf" not in st.session_state:
        st.session_state.ctrl_plots_pdf = None

    if "table_data" not in st.session_state:
        st.session_state.table_data = None

    if "table_data_axis" not in st.session_state:
        st.session_state.table_data_axis = None

    if "figure_cache" not in st.session_state:
        st.session_state.figure_cache = {}


def invalidate_figure_cache():
    """Invalidates the figure cache."""
    st.session_state.figure_cache = {}
    logger.info("invalidated figure cache")


@st.cache_data
def get_brain_regions(raw_files: List[UploadedFile]) -> Iterable[str]:
    """
    Get the brain regions from the uploaded files.

    Args:
        raw_files (List[UploadedFile]): The raw uploaded files to process.

    Returns:
        List[str]: A list of brain regions extracted from the raw files.
    """
    logger.info("get_brain_regions")
    brain_regions = set()
    for raw_file in raw_files:
        _, brain_region, _ = collect_info_from_filename(raw_file.name)
        brain_regions.add(brain_region)
    return list(brain_regions)


def cached_plot_summary_data(
    table_data: pd.DataFrame, project_name: str
) -> Tuple[Figure, BytesIO]:
    """
    Generate a cached plot of summary data.

    Args:
        table_data (pd.DataFrame): The input DataFrame containing the data for the plot.
        project_name (str): The name of the project.

    Returns:
        Tuple[Figure, BytesIO]: A tuple containing the generated plot (Figure object)
            and the BytesIO stream.
    """
    fig = plot_summary_data(table_data, project_name)
    fig_stream = fig2stream(fig)
    return fig, fig_stream


def cached_plot_signal_intensity_along_axis(
    project_name: str, table_data_axis: pd.DataFrame, pixel_size: float
) -> Tuple[Figure, BytesIO]:
    """
    Generate a plot of signal intensity along an axis.

    Args:
        project_name (str): The name of the project.
        table_data_axis (pd.DataFrame): The table data containing the signal intensity
            values along the axis.
        pixel_size (float): The size of each pixel.

    Returns:
        Tuple[Figure, BytesIO]: A tuple containing the generated figure and a BytesIO
            object representing the figure.

    """
    logger.info("creating signal intensity along axis")
    fig = plot_signal_intensity_along_axis(project_name, table_data_axis, pixel_size)
    fig_stream = fig2stream(fig)
    return fig, fig_stream


def get_figure_by_brain_region(
    figures: List[pypdf.PdfWriter], brain_areas: List[str]
) -> Dict[str, List[pypdf.PdfWriter]]:
    """
    Group the given figures by brain region.

    Args:
        figures (List[pypdf.PdfWriter]): A list of figures to be grouped.
        brain_areas (List[str]): A list of brain areas corresponding to each figure.

    Returns:
        Dict[str, List[pypdf.PdfWriter]]: A dictionary where the keys are brain regions
            and the values are lists of figures belonging to each brain region.
    """
    logger.info("get_figure_by_brain_region")
    figures_out = {}
    for fig, brain_area in zip(figures, brain_areas):
        if brain_area not in figures_out:
            figures_out[brain_area] = []

        figures_out[brain_area].append(fig)

    return figures_out


def process_images(
    raw_files: List[UploadedFile],
    pixel_size: float,
    is_masked: bool,
    cache: Dict[Tuple[str, float, bool], Tuple[pypdf.PdfWriter, dict, dict]] = None,
) -> Tuple[List[plt.Figure], pd.DataFrame, pd.DataFrame]:
    """
    Process a list of raw image files.

    Args:
        raw_files (List[UploadedFile]): A list of raw image files.
        pixel_size (float): The pixel size.
        is_masked (bool): A flag indicating whether the images are masked.
        cache (Dict[(str, float, bool), (pypdf.PdfWriter, dict, dict)], optional):
            Use st.session_state.figure_cache as an intermediate cache. Defaults to None

    Returns:
        Tuple[List[plt.Figure], pd.DataFrame, pd.DataFrame]: A tuple containing the
            processed figures, table data, and table data axis.
    """

    if not raw_files:
        return [], None, None

    logger.info("process_images")
    table_data = pd.DataFrame()
    table_data_axis = pd.DataFrame()

    figures = []

    progress_bar = st.progress(0.0, text=f"Processing image {1}/{len(raw_files)}")
    progress_step_size = 1.0 / len(raw_files)

    for i, raw_image in enumerate(raw_files):
        progress_bar.progress(
            i * progress_step_size, text=f"Processing image {(i+1)}/{len(raw_files)}"
        )
        fig, data_row, data_axis_row = process_image_single(
            raw_image, pixel_size, is_masked, cache=cache
        )

        if np.sum(table_data.shape) == 0:
            table_data = pd.DataFrame(columns=data_row.keys())
        table_data.loc[len(table_data)] = data_row

        if np.sum(table_data_axis.shape) == 0:
            table_data_axis = pd.DataFrame(columns=data_axis_row.keys())
        table_data_axis.loc[len(table_data_axis)] = data_axis_row

        figures += [fig]

    # Compute the percentage of white pixels
    table_data["percent_signal"] = (
        table_data["pixels_signal"] / table_data["pixels_total"] * 100
    )

    progress_bar.empty()
    return figures, table_data, table_data_axis


def process_image_single(raw_image, pixel_size, is_masked, cache={}):
    """
    Process a single image.

    Args:
        raw_image (object): The raw image object.
        pixel_size (float): The pixel size.
        is_masked (bool): Flag indicating whether the image is masked.
        cache (dict, optional): The cache dictionary to store processed images.
            Defaults to None.

    Returns:
        tuple: A tuple containing the processed image, temporary variables,
            and temporary axis.
    """

    cache_key = (raw_image.file_id, pixel_size, is_masked)
    if cache_key in cache:
        logger.info(f"process_image_single found cache for {cache_key}")
        return cache[cache_key]

    try:
        animal, brain_area, group = collect_info_from_filename(raw_image.name)
    except ValueError as e:
        logger.error(f"Error: {e}")
        st.warning(f"Error: {e}")
        st.stop()
        return None, None, None
    try:
        fig, _temp_, _temp_axis_ = process_image(
            raw_image,
            is_masked,
            pixel_size,
            animal=animal,
            brain_area=brain_area,
            group=group,
        )
    except UnidentifiedImageError as e:
        logger.error(f"Error: {e}")
        st.warning(f"Error: {e}")
        st.stop()
        return None, None, None

    pdf_fig = fig2pdfpage(fig)
    pdf_fig = pages2pdf([pdf_fig])
    plt.close(fig)

    cache[(raw_image.file_id, pixel_size, is_masked)] = (pdf_fig, _temp_, _temp_axis_)
    logger.info(f"process_image_single created cache for {cache_key}")

    return pdf_fig, _temp_, _temp_axis_



# This file is used to run the streamlit app using
# python -m axoden.gui.streamlit_app

import os

import streamlit.web.bootstrap
from streamlit import config as _config

print("main from axoden.gui.streamlit_app")

dirname = os.path.dirname(__file__)
# filename = os.path.join(dirname, "axoden/gui/streamlit_app/1_📊️_Axoden.py
filename = os.path.join(dirname, "1_📊️_Axoden.py")

_config.set_option("server.headless", True)
args = []

streamlit.web.bootstrap.run(filename, args=[], flag_options={}, is_hello=False)



from io import BytesIO
from typing import List, Union

from matplotlib.figure import Figure
from pypdf import PdfReader, PdfWriter
from pypdf._page import PageObject


def fig2pdfpage(fig: Figure) -> PageObject:
    """Converts a matplotlib figure to a PyPDF PageObject.

    Args:
        fig (Figure): matplotlib figure

    Returns:
        PageObject: PyPDF PageObject containing the input matplotlib figure
    """
    out_pdf = BytesIO()
    fig.savefig(out_pdf, format="pdf")
    out_pdf = PdfReader(out_pdf)
    return out_pdf.pages[0]


def fig2stream(fig: Figure) -> BytesIO:
    """Converts a matplotlib figure to a BytesIO stream.

    The output stream can be directly used as a download in Streamlit.

    Args:
        fig (Figure): matplotlib figure

    Returns:
        BytesIO: BytesIO stream of the input matplotlib figure

    """
    stream = BytesIO()
    fig.savefig(stream, format="pdf")
    return stream


def pdf2stream(pdf: PdfWriter) -> BytesIO:
    """Converts a PyPDF PdfReader to a BytesIO stream.

    The output stream can be directly used as a download in Streamlit.

    Args:
        pdf (PdfWriter): PyPDF PdfWriter object

    Returns:
        BytesIO: BytesIO stream of the input PyPDF PdfReader object
    """
    if not pdf:
        return None

    pdf_stream = BytesIO()
    pdf.write_stream(pdf_stream)
    return pdf_stream


def pages2pdf(pages: List[PageObject]) -> PdfWriter:
    """Converts a list of PyPDF Pages to a single pdf.

    Args:
        pages (List[PageObject]): list of pages

    Returns:
        PdfWriter: PyPDF PdfWriter object containing all input pages
    """
    if not pages:
        return None

    pdf = PdfWriter()
    for page in pages:
        pdf.add_page(page)
    return pdf


def join_pdfs(pdfs: List[Union[PdfReader, PdfWriter]]) -> BytesIO:
    """Joins all first pages of multiple pdfs into a single pdf.

    This function only cares about the first page.
    These types of pdfs are used when figures are converted to pdfs containing only one
    page. The output BytesIO stream can be directly used as a download in Streamlit.

    Args:
        pdfs (List[PdfReader | PdfWriter]): list of pdfs

    Returns:
        BytesIO: BytesIO stream of the joined pdf

    """
    pdf = PdfWriter()
    for p in pdfs:
        page = p.pages[0]
        pdf.add_page(page)
    pdf_stream = BytesIO()
    pdf.write_stream(pdf_stream)
    return pdf_stream



import streamlit as st

st.set_page_config(
    page_title="Cite Axoden",
    page_icon="📖️"
)

def cite_app():
    st.markdown('''
        # Citing AxoDen
        If you use this app in your research, cite it as follows:
        ```
        AxoDen: An Algorithm for Automated Quantification of Axonal Density in defined Brain Regions.
        Raquel Adaia Sandoval Ortega, Emmy Li, Oliver Joseph, Pascal Dufour, Gregory Corder
        BioRxiv
        May 2024
        doi: TODO
        ```
        TODO: add full citation in different formats
    ''')

cite_app()



# Configuration file for the Sphinx documentation builder.
#
# For the full list of built-in configuration values, see the documentation:
# https://www.sphinx-doc.org/en/master/usage/configuration.html

# -- Project information -----------------------------------------------------
# https://www.sphinx-doc.org/en/master/usage/configuration.html#project-information

# Path setup
import os
import sys

sys.path.insert(0, os.path.abspath("../.."))

import axoden

project = "AxoDen"
copyright = "2024, Raquel Adaia Sandoval Ortega, Pascal Dufour"  # noqa A001
author = "Raquel Adaia Sandoval Ortega, Pascal Dufour"
# release = "0.2.1"
release = axoden.__version__

# -- General configuration ---------------------------------------------------
# https://www.sphinx-doc.org/en/master/usage/configuration.html#general-configuration

extensions = [
    "sphinx.ext.napoleon",
    "sphinx.ext.autodoc",
    "sphinx.ext.autosummary",
]

autosummary_generate = True
autosummary_imported_members = True

templates_path = ["_templates"]
exclude_patterns = []


# -- Options for HTML output -------------------------------------------------
# https://www.sphinx-doc.org/en/master/usage/configuration.html#options-for-html-output

html_theme = "alabaster"
html_static_path = ["_static"]



import streamlit as st

from streamlit_pdf_viewer import pdf_viewer


def how_to_app():
    st.title("How To Use AxoDen ")

    st.markdown(
        """
#### Disclaimer
We know that data needs to be handled with care and needs to remain private, especially in a research setting.
Therefore, we designed AxoDen to be a web app that does not store any data on the server.
Data uploaded to the web app is kept in memory for the duration of the analysis.  
The memory is erased when closing or refreshing the browser.  
No third party can access the data in memory.  
The developers of AxoDen have no access to any uploaded nor generated data on the AxoDen web app.  
Only the user can see, interact and store the data the user has uploaded and generated.  
"""
    )

    st.markdown(
        """
## Part 1: Prepare your images
Preparing your images for AxoDen requires the use of 3rd party software.

We recommend using ImageJ or FIJI to modify your images because programs such as Adobe Illustrator
can redistribute the color information into 3 or more channels when saving the modified image as .tif.

Download ImageJ/FIJI here: [imagej.net](https://imagej.net/)

AxoDen is unable to process more than 1 channel containing pixel values above 0.

"""
    )

    st.markdown(
        """
#### Steps

1. Make sure the format of your images is .tif.

2. Rename the files of your images with the following format:  
    *subjectID_brainregion_group_[additinal information if needed].tif.*
    - Do not use spaces in the name of your file.
    - AxoDen interprets the underscore sign as a spacer.
    - AxoDen ignores any information contained in the file name that is beyond the 3rd underscore sign.
    - If you have only one experimental group, you still need to fill the space for group  
    *i.e. "group", "control", "none"...*
    - If you have ROIs of two names, use initials or separate them using a dash sign
    *i.e. "Centrolateral thalamus" can be named as "CL", "TH-CL", "THCL", "centro-lateral-thalamus"...*

3. If you want to mask your images to analyze a precise brain region:
    1. Use Adobe Illustrator or Power Point to overlay the brain atlas to your image and set the boundaries of your brain region of interest.   
    2. Open the image of interest in ImageJ/FIJI.  
    3. Select the freehand selection tool on the top menu.        
    4. Draw the boundaries of the region of interest according to the overlay you created in step 3.1.  
    5. On the top menu select in order:
        - Edit > Clear Outside.
        - Image > Crop.
    6. If the signal to noise ratio needs to be enhanced:
        - On the top menu select Image > Adjust > Brightness/Contrast
        - Decrease the “Maximum” value using the sliding bar to increase fluorescence intensity.
        - Increase the “Minimum” value using the sliding bar to decrease background fluorescence.
    7. Save the resulting image following AxoDen naming convention.

> :point_up: :blue[_TIP:_]  
> _Use the histogram (Analyze > Histogram) to confirm that the background fluorescence in the brain region does not contain zero values._

:warning: TODO: add image of final result in ImageJ/FIJI
"""
    )

    st.markdown("""## Part 2: AxoDen analysis""")

    st.markdown("""#### Steps""")

    st.markdown(
        """
1. Give a name to your project.
2. Change the pixel size to your specification.
The pixel size will depend on your objective and machine you used to acquire the images.

> :point_up: :blue[_TIP:_]  
> _If you are not interested on the ROIs' areas, you don't need to change this value._

3. Keep the check box if your images are masked.
4. Drag and drop all the files you want to analyze.
5. Right after the files have been uploaded, the analysis runs automatically.
6. Next you want to inspect the control plots, see the next section.

"""
    )

    st.markdown("""#### Control Plots""")

    st.image("media/control_plots.jpg", width=400)

    st.markdown(
        """
Control Plots are a set of 4 plots that help you to inspect the quality of your images and how they are processed.
One control plot is created for each image uploaded.

Inspect the control plots to make sure the masking of the image is accurate.
- The mask appears on the second row of the first column.
- The mask is shown in white.
- The mask should cover any part of the image that contains tissue.
- If the mask present black spots within, that is because the background fluorescence of the tissue is 0.
- To correct the mask, repeat Part 1, steps 3.5 to h to create a new masked image.
- If you cannot correct the mask, it means that during acquistion the background fluorescence was too low. **Consider retaking the image.**

Inspect the control plots to make sure the signal does not bleed to surrounding areas.
- In this case, you want to examine the binarized image.
- The binarized image appears on the first row of the second column.
- The signal is shown in white.
- If the signal extends beyond the axon limits (=bleeding to surrounding areas),
it means the signal intensity is too high.
- To correct, repeat Part 1, steps 3.5 to h to create a new masked image.
- If you cannot correct the bleeding, your image is likely overexposed. **Consider retaking the image.**

You can also download the control plots to check them in detail and keep them as reference for later use.
Simply click the “Download figure as pdf” button.
"""
    )

    st.markdown(
        """
#### Summary Data Figure
"""
    )

    pdf_viewer("media/summary_data.pdf", key="summary_data")

    st.markdown(
        """
The Summary Data figure is a collection of 3 plots.
- These are bar plots showing the data as mean +/- SEM.
- Each dot represents one animal/subject.
- The plot on the first column shows the *"% of innervation"* of all ROIs.
These are the final results of AxoDen that you can use for your publication.  
- The plots on the second and third columns are "sanity checks".
- The plot of the second column shows the *"Area"* of each ROI.
You want low variance (= short error bars) for each ROI. Low variance means all selected ROIs have similar sizes.  
- The plot of the third column shows the *"Threshold"* selected with the Otsu method to separate signal from background within the tissue.
Low variances (= short error bars) indicate that images have similar fluorescence intensity values.  

You can download the Summary Data figure by clicking the “Download figure as pdf” button.

#### Summary Data Table
The Summary Data table is a table in tidy format that you can download to run your own statistics.
- "white_pixels" refer to the signal.
- "black_pixels" refer to the background.
- The download file is a .csv file.
- You can open the .csv file in Excel.
- You can copy-paste columns to Prism for further statistical analysis.

To download the summary data as a .csv (comma separated values), which you can later open in Excel or Prism, click the arrow pointing downwards on the top right of the table.
"""
    )

    st.markdown("#### Summary Data Axis Figure")

    pdf_viewer("media/summary_data_axis.pdf", key="summary_data_axis")

    st.markdown(
        """
**If you are interested in the fluorescence intensity projected to the X and the Y axis, read the next steps:**

Understanding the Summary Data Axis plot.
- The grey lines represent individual animals/subjects.
- The blue line and the shade around it represent the mean +/- SEM, respectively.
- Different gray lines may have different sizes.

> :point_up: :blue[TIP:]  
> Rotate cortical brain regions to align to the X or Y axis and investgate the innervation patterns of the different cortical layers.

To download the Summary Data Axis figure, click the “Download figure as pdf” button.

#### Summary Data Axis Table
- This table is not in tidy format.
- The intensity values along the axes are stored as arrays in one cell.
- We recommend to use Python or MATLAB for the statistical analysis of this data.

To download the Summary Data Axis table, click the arrow pointing downwards on the top right of the table.

**All figures are created so that they can be modified in Adobe Illustrator.**
"""
    )


st.set_page_config(page_title="Axoden Tutorial & How To", page_icon="❔️")

how_to_app()



import logging
import os
import sys

import streamlit as st
from streamlit_pdf_viewer import pdf_viewer

if os.getcwd() not in sys.path:
    sys.path.append(os.getcwd())

from axoden.gui.streamlit_app.app_utils import (
    cached_plot_signal_intensity_along_axis,
    cached_plot_summary_data,
    get_brain_regions,
    get_figure_by_brain_region,
    init_session_state,
    invalidate_figure_cache,
    process_images,
)
from axoden.gui.streamlit_app.pdf_utils import join_pdfs, pdf2stream

MAX_IMAGES = 200  # TODO: decide on file upload limit
DEFAULT_PIXEL_SIZE = 0.75521

logger = logging.getLogger(__name__)
logging.basicConfig(level=logging.WARNING)


def axo_den_app():
    """Main Streamlit application for AxoDen."""
    init_session_state()

    st.title("AxoDen")

    with st.container(border=True):
        st.write(
            "If you're new to AxoDen, see the tutorial and learn how to use AxoDen:"
        )
        st.page_link(
            "pages/2_❔️_Tutorial_&_How_To.py",
            label="Tutorial: How To Use AxoDen",
            icon="❔️",
        )

        st.write("If you use the app for your research, cite it using the following:")
        st.page_link("pages/3_📖️_Cite_Axoden.py", label="Cite AxoDen", icon="📖️")

    st.header("Input")
    project_name = st.text_input(label="Project Name", value="AxoDen Analysis")
    pixel_size = st.number_input(
        "Pixel Size (um):",
        value=DEFAULT_PIXEL_SIZE,
        format="%f",
        on_change=invalidate_figure_cache,
        key="pixel_size",
    )  # Set the pixel size
    st.text(
        "Note:\nThe default pixel size is for the 20x Objective of the Keyence BZ-810X "
        "series.\n"
        "Please change it according to the objective used.\n\n4x Objective: 3.77396\n"
        "20x Objective: 0.75521\n"
    )

    raw_files = st.file_uploader(
        "Upload image here. You can add more or remove them later.",
        type=["tif"],
        accept_multiple_files=True,
    )
    if len(raw_files) > MAX_IMAGES:
        st.warning(
            f"This application is limited to using {MAX_IMAGES} images concurrently. "
            "You uploaded {len(raw_files)}, remaining images will not be used in the "
            "analysis!"
        )
        raw_files = raw_files[:MAX_IMAGES]

    is_masked = st.checkbox(
        "Images are masked (desired brain region are cropped out, "
        "backround fluorescence has values above 0,\n"
        "areas of the image lacking tissue have value 0)",
        value=True,
        on_change=invalidate_figure_cache,
        key="is_masked",
    )

    (
        st.session_state.figures,
        st.session_state.table_data,
        st.session_state.table_data_axis,
    ) = process_images(
        raw_files, pixel_size, is_masked, cache=st.session_state.figure_cache
    )

    # plot table data results
    if st.session_state.table_data is not None:
        logger.info("Creating data section")
        st.header("Summary Data")

        fig, fig_stream = cached_plot_summary_data(
            st.session_state.table_data, project_name
        )

        st.pyplot(fig)
        st.download_button(
            "Download figure as pdf", fig_stream, "axoden_summary_data.pdf"
        )
        st.dataframe(st.session_state.table_data)

    # plot table data by axis results
    if st.session_state.table_data_axis is not None:
        logger.info("Creating data axis section")
        st.header("Summary Data Axis")

        fig, fig_stream = cached_plot_signal_intensity_along_axis(
            project_name, st.session_state.table_data_axis, pixel_size
        )
        st.pyplot(fig)

        st.download_button(
            "Download figure as pdf", fig_stream, "axoden_summary_data_axis.pdf"
        )
        st.dataframe(st.session_state.table_data_axis)

    logger.info("Creating control plots pdf")
    st.session_state.ctrl_plots_pdf = join_pdfs(st.session_state.figures)

    brain_regions = get_brain_regions(raw_files)
    if brain_regions:
        logger.info("Creating control plots by brain area")
        st.header("Control Plots by Brain Area")

        tabs = st.tabs(brain_regions)

        brain_areas = st.session_state.table_data.brain_area.to_list()
        figure_dict = get_figure_by_brain_region(st.session_state.figures, brain_areas)

        for tab, brain_region in zip(tabs, brain_regions):
            figures = figure_dict[brain_region]

            with tab:
                if len(figures) > 1:
                    indices = [str(x) for x in range(len(figures))]
                    tabs_brain_region = st.tabs(indices)
                    for i, tab_fig_nr in enumerate(tabs_brain_region):
                        with tab_fig_nr:
                            pdf_figure = pdf2stream(figures[i]).getvalue()
                            pdf_viewer(pdf_figure, key=f"{brain_region}_{i}")
                else:
                    pdf_figure = pdf2stream(figures[0]).getvalue()
                    pdf_viewer(pdf_figure, key=brain_region)

        st.download_button(
            "Download plots as pdf",
            st.session_state.ctrl_plots_pdf,
            "control_plots.pdf",
        )


if __name__ == "__main__":
    st.set_page_config(page_title="Axoden App", page_icon="📊️")
    axo_den_app()



import os
import re
from pathlib import Path
from typing import Tuple

# this allows pyinstaller on github actions to detect pdf backend
import matplotlib.backends.backend_pdf  # noqa: F401
import matplotlib.gridspec as gridspec
import matplotlib.pyplot as plt
import numpy as np
import pandas as pd
import seaborn as sns
from PIL import Image
from skimage.filters import threshold_otsu
from tqdm import tqdm


def intensity_along_axis(img, ax=None):
    """Compute the intensity along the x or y axis of the image."""
    if ax == "x":
        s = np.nansum(img, axis=0)  # Sum along the x axis
    elif ax == "y":
        s = np.nansum(img, axis=1)  # Sum along the y axis
    elif ax is None:
        s = (
            np.nansum(img, axis=0),
            np.nansum(img, axis=1),
        )  # Sum along the x and y axis
    else:
        raise ValueError(
            f"Expected argument ax to be one of 'x' or 'y' or None, but received '{ax}'"
        )
    return s


def remove_spines_plot(ax, loc=["all"]):
    """Remove the spines of the plot."""
    if len(loc) == 1 and loc[0] == "all":
        loc = ["top", "right", "bottom", "left"]
        [ax.spines[spine_location].set_visible(False) for spine_location in loc]
    else:
        [ax.spines[spine_location].set_visible(False) for spine_location in loc]


def remove_ticks_plot(ax, loc="all"):
    """Remove the ticks of the plot."""
    if loc == "all":
        ax.set_xticks([])
        ax.set_yticks([])
    elif loc == "x":
        ax.set_xticks([])
    elif loc == "y":
        ax.set_yticks([])
    else:
        raise ValueError(
            "Expected argument 'loc' to be one of 'all', 'x', 'y', "
            f"but received '{loc}'"
        )


def plot_intensity_along_axis(ax, x, y, loc):
    if loc == "y":
        ax.fill_betweenx(x, y, color="grey", alpha=0.5)
    elif loc == "x":
        ax.fill_between(x, y, color="grey", alpha=0.5)
    ax.margins(0)


def control_plot_steps(ax, img, title, axs_info):
    """
    Create the subplots of the control plots for each input image.

    Args:
        ax (list of matplotlib Axes): List of Axes objects to plot the images.
        img (numpy array): Input image data.
        title (str): Title for the plot.
        axs_info (dict or list): Information about the axes.

    Returns:
        ax (list of matplotlib Axes): List of Axes objects with the plotted images.
    """

    # Compute x and y axis signals
    x_ax_signal, y_ax_signal = intensity_along_axis(img)

    # Define the label of the quantification plots
    if title != "Mask":
        label_quant = "Signal\nintensity"
    else:
        label_quant = "Dimensions\nmask"

    # Plot original image, binarized image and area of the original image
    ax[0].imshow(img, cmap="gray")
    ax[0].set_title(title, pad=12)
    # Beautify
    remove_ticks_plot(ax[0], loc="x")
    remove_spines_plot(ax[0], loc=["top", "right", "bottom"])

    # Plot the signal in the y axis
    plot_intensity_along_axis(
        ax[1], np.arange(len(y_ax_signal)), y_ax_signal[::-1], "y"
    )
    # Set x label at given location
    ax[1].set_xlabel(label_quant, loc="center", rotation=45, color="grey")
    ax[1].xaxis.set_label_coords(1, -0.01)
    # Beautify
    remove_ticks_plot(ax[1])
    remove_spines_plot(ax[1])

    # Plot the signal in the x axis
    plot_intensity_along_axis(ax[2], np.arange(len(x_ax_signal)), -x_ax_signal, "x")
    # Beautify
    remove_ticks_plot(ax[2], loc="y")
    remove_spines_plot(ax[2], loc=["top", "right", "left"])

    # Match the size of ax[1] and ax[2] to ax[0]
    ax0_size = ax[0].get_position()
    ax1_size = ax[1].get_position()
    ax[1].set_position(
        [ax0_size.x0 + ax0_size.width, ax0_size.y0, ax1_size.width, ax0_size.height]
    )
    ax2_size = ax[2].get_position()
    ax[2].set_position(
        [ax0_size.x0, ax0_size.y0 - ax2_size.height, ax0_size.width, ax2_size.height]
    )

    # Set x ticks and labels for the medio-lateral axis
    x_ax_ticks = axs_info[0]
    x_ax_labels = axs_info[1]
    ax[2].set_xticks(x_ax_ticks)
    ax[2].set_xticklabels(x_ax_labels, rotation=45)
    ax[2].set_xlabel("Medio-Lateral axis (\u03bcm)")

    # Set y ticks and labels for the dorso-ventral axis
    y_ax_ticks = axs_info[2]
    y_ax_labels = axs_info[3]
    ax[0].set_yticks(y_ax_ticks)
    ax[0].set_yticklabels(y_ax_labels)
    ax[0].set_ylabel("Dorso-Ventral axis (\u03bcm)")

    return ax


def get_tif_files(folder_path):
    """Get all the tif files in the folder provided."""
    tif_files = []
    for file in os.listdir(folder_path):
        if file.endswith(".tif"):
            tif_files.append(os.path.join(folder_path, file))
    return tif_files


def collect_info_from_filename(filename: str) -> Tuple[str, str, str]:
    """Collect the animal, brain area and group from the filename.

    Args:
        filename (str): Name of the file.

    Returns:
        Tuple[str, str, str]: A tuple containing the animal, brain area and group.
    """
    img_name = os.path.basename(filename)
    img_name, _ = os.path.splitext(img_name)

    # img_name = img_name.split('.')[0]
    name_parts = img_name.split("_")
    if len(name_parts) < 2:
        raise ValueError(
            "Expected filename to include at least one '_' character to split animal "
            "from brain area, e.g. 'animal_brainarea.tif', "
            f"but received the file name '{filename}'"
        )
    animal = name_parts[0]
    brain_area = name_parts[1]

    # backward compatibility for only animal and brain area
    group = "group1"
    if len(name_parts) >= 3:
        group = name_parts[2]

    return animal, brain_area, group


def open_tif_image(img_path):
    """Open a tif image.

    Args:
        img_path: A filename (string), os.PathLike object or a file object.

    Returns:
        Image: A PIL Image object.
    """
    return Image.open(img_path)  # works well also with Uploaded image


def generate_background_mask(img, is_masked):
    """Generate a background mask for the image provided."""

    image_matrix = np.array(img)

    # For non-masked images
    if not is_masked:
        mask = np.zeros(image_matrix.shape[0:2], dtype="bool")

    # For masked images
    else:
        if image_matrix.ndim == 3:
            mask = np.all(image_matrix == [0, 0, 0], axis=-1)
        elif image_matrix.ndim == 2:
            mask = image_matrix == 0
        else:
            raise ValueError(
                "Expected image of dimension 2 or 3, "
                f"but reiceived shape {image_matrix.shape}"
            )

    return mask


def convert_image_to_gray(img):
    return img.convert("L")


def collect_within_mask(img, mask):
    return np.where(mask, np.nan, img)


def collect_image_mask(img_path: any, is_masked: bool) -> Tuple[np.ndarray, np.ndarray]:
    """Open the image and cut it to the area that contains tissue.

    Args:
        img_path: A filename (string), os.PathLike object or a file object.
        is_masked: True if the image was masked and contains regions that are set
            to a value of 0.

    Returns:
        Tuple[np.ndarray, np.ndarray]: A tuple containing the cut image and the mask.
    """
    img = open_tif_image(img_path)
    # Find the area of the image that cotains tissue
    mask = generate_background_mask(img, is_masked)
    # Convert image to grayscale
    image_gray = convert_image_to_gray(img)
    # Cut the image to the area that contains tissue
    cut_image = collect_within_mask(image_gray, mask)
    return cut_image, mask


def binarize_image(image, threshold):
    """Binarize the image using the threshold provided."""
    binary_image = np.where(image > threshold, 1, 0)
    return binary_image


def count_pixels(img):
    """Count the number of white and black pixels in the image."""

    # make sure the image is binarized and only contains black (0) and white (1) pixels
    # otherwise this function will not behave properly!
    unique_values = np.unique(img)
    if not (len(unique_values) == 2 and np.all(np.unique(img) == [0, 1])):
        raise ValueError(
            "Expected a binarized image with only zeros and ones, "
            f"but received values {unique_values}"
        )

    white_pixels = np.sum(img)
    black_pixels = np.sum(1 - img)
    all_pixels = white_pixels + black_pixels
    return [white_pixels, black_pixels, all_pixels]


def compute_area(img, pixel_size):
    """Compute the area of the image and the area of the white and black pixels."""
    counts = count_pixels(img)
    [w, b, total] = [c * (pixel_size**2) for c in counts]
    return w, b, total


def compute_threshold(img):
    non_nan_values = img[~np.isnan(img)]
    thr = threshold_otsu(np.array(non_nan_values))
    # If the threshold is too low, set it to 20
    if thr <= 10:
        thr = np.int32(20)
    return thr


def make_figures_pdf_editable():
    """Make the figures editable in Adobe Illustrator."""
    plt.rcParams["pdf.fonttype"] = 42
    plt.rcParams["ps.fonttype"] = 42
    plt.rcParams["font.family"] = "sans-serif"
    plt.rcParams["font.sans-serif"] = "DejaVu Sans"


def prepare_axis_information(img, pixel_size):
    x_ax_pixels = np.linspace(0, img.shape[1], 6)
    x_ax_um = np.round(x_ax_pixels * pixel_size, decimals=1)
    y_ax_pixels = np.linspace(0, img.shape[0], 6)
    y_ax_um = np.round(y_ax_pixels * pixel_size, decimals=1)
    return [x_ax_pixels, x_ax_um, y_ax_pixels, y_ax_um]


def load_table(file_path: str):
    """Load a table from a csv file.

    Args:
        file_path (str): Path to the csv file.

    Returns:
        pd.DataFrame: A DataFrame with the data from the csv file.
    """

    def _string_to_ndarray(string_list):
        """Convert a string representation of a list of numbers to a numpy array."""
        return np.array([int(n) for n in re.findall(r"\d+", string_list)])

    col_types = {"animal": str, "brain_area": str}
    table_data = pd.read_csv(file_path, dtype=col_types, index_col=False)

    # Remove any unnamed columns. If the .csv was saved from the streamlit app,
    # it will contain an unnamed index column
    table_data = table_data.loc[:, ~table_data.columns.str.contains("^Unnamed")]

    for col in table_data.columns:
        if "signal_bin" in col or "signal_gray" in col:  # those are ndarray columns
            table_data[col] = table_data[col].apply(_string_to_ndarray)

    return table_data


def save_table(df: pd.DataFrame, folderpath: str, filename: str):
    """Save a table to a csv file.

    Args:
        df (pd.DataFrame): DataFrame to save.
        folderpath (str): Path to the folder where the file will be saved.
        filename (str): Name of the file.
    """
    df = df.copy()  # make sure the original dataframe is not modified

    Path(folderpath).mkdir(parents=True, exist_ok=True)

    filepath = os.path.join(folderpath, filename)
    if os.path.exists(filepath):
        os.remove(filepath)

    # convert numpy arrays to lists, otherwise the string representation is saved
    # as [0, 5, 2, ..., 3, 4, 5]
    for col in df.columns:
        if isinstance(df[col][0], np.ndarray):
            # [1:-1] removes brackets
            # .replace(' ', '') removes spaces to make file size smaller
            df[col] = df[col].apply(
                lambda x: str(list(x.astype(int)))[1:-1].replace(" ", "")
            )
    df.to_csv(filepath, index=False)


def generate_control_plot(img, img_bin, msk, pixel_size, info_pie):
    # Create a grid for the subplots
    gs = gridspec.GridSpec(13, 11)

    # Prepare axis informationTransform x and y axes from pixels to mm
    axis_info = prepare_axis_information(img, pixel_size)

    # Subplot original image
    ax_ori = [
        plt.subplot(gs[0:4, 0:4]),
        plt.subplot(gs[0:4, 4]),
        plt.subplot(gs[4, 0:4]),
    ]
    ax_ori = control_plot_steps(ax_ori, img, "Grayscale image", axis_info)

    # Subplot binarized image
    ax_bin = [
        plt.subplot(gs[0:4, 6:10]),
        plt.subplot(gs[0:4, 10]),
        plt.subplot(gs[4, 6:10]),
    ]
    ax_bin = control_plot_steps(ax_bin, img_bin, "Binarized image", axis_info)

    # Subplot area of the original image
    ax_msk = [
        plt.subplot(gs[7:11, 0:4]),
        plt.subplot(gs[7:11, 4]),
        plt.subplot(gs[11, 0:4]),
    ]
    ax_msk = control_plot_steps(ax_msk, ~msk, "Mask", axis_info)

    # Subplot pie chart
    ax_pie = plt.subplot(gs[7:11, 6:11])
    ax_pie.pie(
        info_pie["sizes"],
        labels=info_pie["labels"],
        autopct="%1.1f%%",
        startangle=90,
        colors=info_pie["colors"],
        labeldistance=1.2,
        wedgeprops={"edgecolor": "black", "linewidth": 0.5, "antialiased": True},
    )
    # set title to subplot
    ax_pie.set_title("Proportion of signal vs background", pad=12)
    # Set aspect ratio to equal for a circular pie chart
    ax_pie.set_aspect("equal")
    # Add a line to the pie chart
    ax_pie.add_line(plt.Line2D([0.5, 0.5], [0.5, 0.5], color="black", linewidth=1))


def process_image(
    file_name: any,
    is_masked: bool,
    pixel_size: float,
    animal: str = "animal1",
    brain_area: str = "brain_area1",
    group: str = "group1",
) -> Tuple[plt.Figure, dict, dict]:
    """Process a single image and generate a control plot.

    Args:
        file_name (any): A filename (string), os.PathLike object or a file object.
        is_masked (bool): True if the image was masked and contains regions that are
            set to a value of 0.
        pixel_size (float): Pixel size in micrometers.
        animal (str): Name of the animal. Default is "animal1".
        brain_area (str): Name of the brain area. Default is "brain_area1".
        group (str): Name of the group. Default is "group1".

    Returns:
        Tuple[plt.Figure, dict, dict]: A tuple containing the figure, \
            the data for the image and the data with the axis projections.
    """
    img, msk = collect_image_mask(file_name, is_masked)
    msk_bool = msk.astype(bool)
    thr = compute_threshold(img[~msk_bool])
    img_bin = binarize_image(img, thr)

    [w, b, total] = count_pixels(img_bin[~msk_bool])
    area_w, area_b, area_img = compute_area(img_bin[~msk_bool], pixel_size)
    area_image_mm = area_img / 1000

    # Append the information to the DataFrame for the image
    data = {
        "animal": animal,
        "brain_area": brain_area,
        "group": group,
        "pixels_signal": w,
        "pixels_black": b,
        "pixels_total": total,
        "threshold": thr,
        "area_image": area_img,
        "area_signal": area_w,
        "area_black": area_b,
        "area_img_mm": area_image_mm,
    }

    # Append the information to the DataFrame for the axis
    axis_data = {
        "animal": animal,
        "brain_area": brain_area,
        "group": group,
        "signal_bin_x_ax": intensity_along_axis(img_bin, "x"),
        "signal_bin_y_ax": intensity_along_axis(img_bin, "y"),
        "signal_gray_x_ax": intensity_along_axis(img, "x"),
        "signal_gray_y_ax": intensity_along_axis(img, "y"),
    }

    # Generate control plot
    fig = plt.figure(figsize=(8, 8))
    fig.suptitle(
        f"Animal {animal} | {brain_area} | Area: {area_image_mm:.2f}mm\u00b2| "
        f"{group} | Threshold: {thr:.2f}",
        weight="bold",
    )
    info_pie = {
        "labels": ["Area receiving\nprojections", "Area without\nprojections"],
        "sizes": [area_w, area_b],
        "colors": ["white", "grey"],
    }
    generate_control_plot(img, img_bin, msk, pixel_size, info_pie)

    return fig, data, axis_data


def process_folder(
    folder_path: str,
    pixel_size: float,
    is_masked: bool,
    output_folder: str = None,
    save: bool = True,
) -> Tuple[pd.DataFrame, pd.DataFrame]:
    """Process all images in a folder.

    Optionally save the data to csv files and the control plots to pdf files.

    Args:
        folder_path (str): Path to the folder containing the images.
        pixel_size (float): Pixel size in micrometers.
        is_masked (bool): True if the image was masked and contains regions that are
            set to a value of 0.
        output_folder (str): Path to the folder where the data will be saved.
            If not set, the input folder will be used.
        save (bool): True if the control plots should be saved. Default is True.

    Returns:
        Tuple[pd.DataFrame, pd.DataFrame]: A tuple containing the data for the images \
            and the data for the axis.
    """

    if not output_folder:
        output_folder = folder_path

    # Set fonts editable in Adobe Illustrator
    make_figures_pdf_editable()

    # Create an empty DataFrame to store the information
    table_data = pd.DataFrame()
    table_data_axis = pd.DataFrame()

    # Create a folder for the control plots if it doesn't exist in the folder path
    control_plot_path = os.path.join(output_folder, "control_plots")
    if save:
        Path(control_plot_path).mkdir(parents=True, exist_ok=True)

    # Create a list with all the tif files in the folder
    file_list = get_tif_files(folder_path)
    if len(file_list) == 0:
        raise ValueError("No tif files found in the folder path")

    # Loop through all the images in the folder
    for filepath in tqdm(file_list, desc="Processing images"):
        animal, brain_area, group = collect_info_from_filename(filepath)
        fig, _temp_, _temp_axis_ = process_image(
            filepath,
            is_masked=is_masked,
            pixel_size=pixel_size,
            animal=animal,
            brain_area=brain_area,
            group=group,
        )

        if np.sum(table_data.shape) == 0:
            table_data = pd.DataFrame(columns=_temp_.keys())
        table_data.loc[len(table_data)] = _temp_

        if np.sum(table_data_axis.shape) == 0:
            table_data_axis = pd.DataFrame(columns=_temp_axis_.keys())
        table_data_axis.loc[len(table_data_axis)] = _temp_axis_

        # Save control plot
        if save:
            figure_name = f"{animal}_{brain_area}_control_plot.pdf"
            figure_path = os.path.join(control_plot_path, figure_name)
            plt.savefig(figure_path, dpi=300)
        plt.close(fig)

    # Compute the percentage of white pixels
    table_data["percent_signal"] = (
        table_data["pixels_signal"] / table_data["pixels_total"] * 100
    )

    # Save dataframes as csv files
    if save:
        save_table(table_data, output_folder, "projections_quantification.csv")
        save_table(
            table_data_axis, output_folder, "projections_quantification_axis.csv"
        )

    return table_data, table_data_axis


def write_summary_data_plot(
    folder_path: str, df_input: pd.DataFrame, project_name: str = None
):
    """Write the summary data plot to a pdf file in the folder provided.

    Args:
        folder_path (str): Path to the folder where the data will be saved.
        df_input (pd.DataFrame): Dataframe with the data to plot.
        project_name (str): Name of the project. If not set, default to the folder name.
    """

    Path(folder_path).mkdir(parents=True, exist_ok=True)

    if not project_name:
        project_name = os.path.basename(folder_path)

    fig = plot_summary_data(df_input, project_name)

    # Save the figure
    figure_name = "projections_quantification.pdf"
    figure_path = os.path.join(folder_path, figure_name)
    fig.savefig(figure_path, dpi=300)
    plt.close()


def plot_summary_data(df_input, project_name):
    # Set fonts editable in Adobe Illustrator
    make_figures_pdf_editable()

    # Create a figure with four subplots
    fig, ax = plt.subplots(1, 4, figsize=(18, 5))
    brain_areas = df_input["brain_area"].unique()
    fig.suptitle(f'Project "{project_name}"\nData Summary', weight="bold")

    # Set the color for the plots
    color_bars = "royalblue"
    color_scatter_in = "white"

    # Create a dictionary with the animals per brain area
    dict_animals_brain_area = {}

    # Define which columns to use for the plot
    data_cols = ["percent_signal", "area_img_mm", "threshold"]
    ylabel_dict = {
        "percent_signal": "Innervation (%)",
        "area_img_mm": "Area (mm\u00b2)",
        "threshold": "Pixel value",
    }
    sublplot_titles = [
        "Percentage of innervation",
        "Area of the brain region",
        "Threshold for binarization",
    ]

    # Plot the bar plot with the mean and sem of the white pixels percentage
    for i_ba, br_ar in enumerate(brain_areas):

        # Get the data for the brain area
        data = df_input[df_input["brain_area"] == br_ar]

        # Iterate over the columns to plot
        for i_col, col in enumerate(data_cols):
            vals = data[col]
            vals_mean = np.mean(vals)
            sem = np.std(vals) / np.sqrt(len(vals))

            # Plot the bar plot with the mean and sem
            ax[i_col].bar(i_ba, vals_mean, color=color_bars, yerr=sem)

            # Plot the scatter plot with round markers
            ax[i_col].scatter(
                np.repeat(i_ba, len(vals)),
                vals,
                edgecolors=color_bars,
                facecolors=color_scatter_in,
            )

        # Fill the dictionary with the animals in the brain area
        dict_animals_brain_area[br_ar] = data["animal"].unique().tolist()

    # Beautify data subplots
    for i_col, col in enumerate(data_cols):
        # Set the column name as the title
        ax[i_col].set_title(sublplot_titles[i_col], weight="bold")
        # Set the xticks and labels for each data column
        ax[i_col].set_xticks(np.arange(0, len(brain_areas)))
        ax[i_col].set_xticklabels(brain_areas, rotation=45)
        # Set the y maximum value
        yval_max = np.ceil(ax[i_col].get_ylim()[1] / 10) * 10
        ax[i_col].set_yticks(np.linspace(0, yval_max, 6))
        ax[i_col].set_ylim(0, yval_max)
        # Set the y label
        ax[i_col].set_ylabel(ylabel_dict[col])

    # Print the number of animals per brain area
    df = pd.DataFrame.from_dict(dict_animals_brain_area, orient="index")
    df["animals"] = df.apply(lambda row: ", ".join(row.astype(str)), axis=1)
    df["animals"] = df["animals"].apply(
        lambda x: x.replace(", None", "")
    )  # Remove None values
    i_anim_col = df.columns.get_loc("animals")
    df = df.iloc[:, i_anim_col:]  # Remove columns 0 to the one with all the animals
    df["brain_area_with_animals"] = (
        df.index + " (N=" + df["animals"].apply(lambda x: str(x.count(",") + 1)) + ")"
    )  # Count the number of animals
    df = df.set_index(
        df["brain_area_with_animals"]
    )  # Replace the index with the values in the new column
    df = df.drop(columns=["brain_area_with_animals"])  # Remove the column
    df.columns = [""] * len(df.columns)  # Remove column
    df.index.name = ""  # Remove index names
    ax[-1].set_title("Animals per Brain Region", weight="bold")
    ax[-1].axis("off")
    ax[-1].text(0.05, 0.1, df.to_string(header=False), fontsize=9)

    # Set the layout & despine
    plt.tight_layout()
    sns.despine(fig=fig, top=True, right=True)

    # Delete the DataFrame
    del df, df_input
    return fig


def write_signal_intensity_along_axis_plot(
    output_folder: str, df: pd.DataFrame, pixel_size: float, project_name: str = None
):
    """Write the signal intensity along the axis to a pdf file in the folder provided.

    Args:
        folderpath (str): Path to the folder where the data will be saved.
        df (pd.DataFrame): Dataframe with the axis data to plot.
        pixel_size (float): Pixel size in micrometers.
        project_name (str): Name of the project.
            If not set, it will be the name of the output folder.
    """

    # Collect project name
    if not project_name:
        project_name = os.path.basename(output_folder)

    fig = plot_signal_intensity_along_axis(project_name, df, pixel_size)

    # Save the figure
    figure_name = "projections_quantification_along_axis.pdf"
    figure_path = os.path.join(output_folder, figure_name)
    if os.path.exists(figure_path):
        os.remove(figure_path)
    fig.savefig(figure_path, dpi=300)
    plt.close()


def plot_signal_intensity_along_axis(
    project_name: str, df: pd.DataFrame, pixel_size: float
):
    """Plot the signal intensity along the x and y axis of the DataFrame.

    Args:
        project_name (str): Name of the project.
        df (pd.DataFrame): DataFrame with the data to plot.
        pixel_size (float): Pixel size in micrometers.

    Returns:
        plt.Figure: A figure with the signal intensity along the x and y axis.
    """

    # Set fonts editable in Adobe Illustrator
    make_figures_pdf_editable()

    # Get the unique brain areas
    brain_areas = df["brain_area"].unique()
    n_ba = len(brain_areas)

    if n_ba > 5:
        n_y_ticks = 2
        size_tick_lbl = 8
    else:
        n_y_ticks = 4
        size_tick_lbl = 10

    # Define which columns to use for the plot
    data_cols = ["signal_bin_x_ax", "signal_bin_y_ax"]
    subplot_titles = ["Medio-Lateral axis", "Dorso-Ventral axis"]

    # Create a figure with four subplots
    fig, ax = plt.subplots(n_ba, len(data_cols), figsize=(16, 9), squeeze=False)
    fig.suptitle(
        f'Project "{project_name}"\n'
        "Signal Intensity Along the Medio-Lateral and Dorso-Ventral Axes",
        weight="bold",
    )

    # Set the color for the plots
    color_bars = "royalblue"

    # Plot the bar plot with the mean and sem of the white pixels percentage
    for i_ba, br_ar in enumerate(brain_areas):

        # Get the data for the brain area
        data = df[df["brain_area"] == br_ar]

        # Iterate over the columns to plot
        for i_col, col in enumerate(data_cols):

            # Collect data arrays
            vals = data[col].tolist()
            n_vals = len(vals)
            # Pad the shortest array(s) with nans
            max_len = np.max([len(v) for v in vals])
            val_eq_size = []
            for v in vals:
                if len(v) < max_len:
                    val_out = np.concatenate([v, np.zeros(max_len - len(v)) + np.nan])
                else:
                    val_out = v
                val_eq_size.append(val_out)
            vals = np.array(val_eq_size)
            del val_eq_size
            # Compute the mean and sem
            vals_mean = np.nanmean(vals, axis=0)
            sem = np.nanstd(vals, axis=0) / np.sqrt(n_vals)

            # Generate the x axis values
            n_pts = vals_mean.shape[0]
            x_ax = np.arange(0, n_pts, 1)
            x_ax_ticks = np.round(np.linspace(0, n_pts, 10), 1)
            x_ax_labels = np.round(x_ax_ticks * pixel_size, 1)
            # Plot the values of each animal
            ax[i_ba, i_col].plot(
                x_ax, vals.T, color="grey", alpha=0.5, linewidth=0.7
            )  # FIXME: crashes with only a single animal, need to test
            # Plot the bar plot with the mean and sem
            ax[i_ba, i_col].plot(x_ax, vals_mean, color=color_bars, linewidth=1.5)
            ax[i_ba, i_col].fill_between(
                x_ax, vals_mean - sem, vals_mean + sem, color=color_bars, alpha=0.2
            )

            # Beautify data subplots
            ax[i_ba, i_col].margins(0)
            ax[i_ba, i_col].set_xticks(x_ax_ticks)
            ax[i_ba, i_col].set_xticklabels(x_ax_labels, fontsize=size_tick_lbl)
            # Set the x label at the bottom
            if i_ba == n_ba - 1:
                ax[i_ba, i_col].set_xlabel(
                    f"{subplot_titles[i_col]} (\u03bcm)", weight="bold"
                )
            # Name the y axis as the brain area
            if i_col == 0:
                ax[i_ba, i_col].set_ylabel(
                    f"{br_ar}", weight="bold", rotation=0, labelpad=20, ha="right"
                )
            # Set the y maximum value
            yval_max = np.ceil(ax[i_ba, i_col].get_ylim()[1] / 10) * 10
            yticks = np.linspace(0, yval_max, n_y_ticks)
            ax[i_ba, i_col].set_yticks(yticks)
            ax[i_ba, i_col].set_yticklabels(yticks, fontsize=size_tick_lbl)
            ax[i_ba, i_col].set_ylim(0, yval_max)

    # Set the layout & despine
    sns.despine(fig=fig, top=True, right=True)
    plt.subplots_adjust(top=0.92, bottom=0.05, hspace=0.5, wspace=0.2)

    return fig


if __name__ == "__main__":
    pixel_size = (
        0.75521  # um, based on 20x objective from Keyence BZ-810X series microscope.
    )
    input_dir = "sample_images"
    output_dir = "output_data"
    is_masked = True

    # Collect the data and save it as a csv file
    quant_projections, quant_along_axis = process_folder(
        input_dir, pixel_size, is_masked, output_folder=output_dir
    )

    # Plot the summary data
    write_summary_data_plot(output_dir, quant_projections)

    # Plot the average of the signal intensity along the x and y axis
    write_signal_intensity_along_axis_plot(output_dir, quant_along_axis, pixel_size)
